from datasets import load_dataset
import string


def get_labels():
    return ["person", "organization", "location", "miscellaneous"]


def get_annotated_sentences():
    def join_tokens(tokens):  # Helper function to join tokens into a well-formatted sentence.
        sentence = ""
        for token in tokens:
            # If token is punctuation, append it without a preceding space.
            if token in string.punctuation:
                sentence += token
            else:
                # Otherwise, add a space (if not the first token) then the token.
                sentence += (" " if sentence else "") + token
        return sentence

    # Load the training split from the CoNLL2003 dataset.
    print("Loading dataset: CoNLL2003")
    dataset = load_dataset("eriktks/conll2003", split="train", trust_remote_code=True)
    print("Finished loading dataset: CoNLL2003")

    print("Process dataset: CoNLL2003")
    # Mapping from the raw entity type (from the dataset) to the desired label.
    label_mapping = {
        "PER": "person",
        "ORG": "organization",
        "LOC": "location",
        "MISC": "miscellaneous"
    }

    # Prepare a list to hold the annotated sentences.
    annotated_sentences = []

    # The dataset's ner_tags are stored as integers. Use the int2str method to get string tags.
    int2str = dataset.features["ner_tags"].feature.int2str

    # Iterate over each example in the training dataset.
    for example in dataset:
        tokens = example["tokens"]
        # Convert numerical NER tags to their string representation.
        ner_tags = [int2str(tag) for tag in example["ner_tags"]]
        # Reconstruct the full sentence.
        sentence = join_tokens(tokens)

        # Extract entities from the sentence using the IOB tagging scheme.
        entities = []
        current_tokens = []
        current_label = None

        for token, tag in zip(tokens, ner_tags):
            if tag == "O":
                if current_tokens:
                    # End of an entity; join tokens and append the entity tuple.
                    entity_text = join_tokens(current_tokens)
                    entities.append((entity_text, label_mapping.get(current_label, current_label)))
                    current_tokens = []
                    current_label = None
            else:
                # The tag is in the format B-XXX or I-XXX.
                tag_prefix, tag_type = tag.split("-", maxsplit=1)
                if tag_prefix == "B":
                    # If an entity was in progress, finish it.
                    if current_tokens:
                        entity_text = join_tokens(current_tokens)
                        entities.append((entity_text, label_mapping.get(current_label, current_label)))
                    # Start a new entity.
                    current_tokens = [token]
                    current_label = tag_type
                elif tag_prefix == "I":
                    if current_tokens and current_label == tag_type:
                        # Continue the current entity.
                        current_tokens.append(token)
                    else:
                        # In case of an I-tag without a preceding B-tag, start a new entity.
                        if current_tokens:
                            entity_text = join_tokens(current_tokens)
                            entities.append((entity_text, label_mapping.get(current_label, current_label)))
                        current_tokens = [token]
                        current_label = tag_type

        # If the sentence ends while an entity is still being processed, add it.
        if current_tokens:
            entity_text = join_tokens(current_tokens)
            entities.append((entity_text, label_mapping.get(current_label, current_label)))

        # Append the annotated sentence.
        # Each entry is a list: [ [sentence], [list of (entity, label) tuples] ]
        annotated_sentences.append([sentence, entities])

    print("Finished processing dataset: CoNLL2003")

    return annotated_sentences
