import data.Figer.entity_pb2 as entity_pb2
from google.protobuf.internal.decoder import _DecodeVarint32
import os
import requests
import gzip
import shutil
import pickle

base_dir = os.path.dirname(os.path.realpath(__file__))

coarse_labels = ["person", "organization", "location", "product", "art", "event", "building"]
fine_labels = ["actor", "architect", "artist", "athlete", "author", "coach", "director", "doctor", "engineer",
               "monarch", "musician", "politician", "religious leader", "soldier", "terrorist", "airline", "company",
               "educational institution", "fraternity sorority", "sports league", "sports team",
               "terrorist organization", "government agency", "government", "political party", "educational department",
               "military", "news agency", "city", "country", "county", "province", "railway", "road", "bridge",
               "body of water", "island", "mountain", "glacier", "astral body", "cemetery", "park", "engine",
               "airplane", "car", "ship", "spacecraft", "train", "camera", "mobile phone", "computer", "software",
               "game", "instrument", "weapon", "film", "play", "written work", "newspaper", "music", "attack",
               "election", "protest", "military conflict", "natural disaster", "sports event", "terrorist attack",
               "airport", "dam", "hospital", "hotel", "library", "power station", "restaurant", "sports facility",
               "theater", "point in time", "color", "award", "educational degree", "title", "law", "ethnicity",
               "language", "religion", "god", "chemical thing", "biological thing", "medical treatment", "disease",
               "symptom", "drug", "body part", "living thing", "animal", "food", "website", "broadcast network",
               "broadcast program", "tv channel", "currency", "stock exchange", "algorithm", "programming language",
               "transit system", "transit line"]

entries_coarse = []
entries_fine = []


def download_and_unpack():
    gzip_path = os.path.join(base_dir, "train.data.gz")
    data_path = os.path.join(base_dir, "train.data")

    if not os.path.exists(data_path):

        if not os.path.exists(gzip_path):
            DATASET_URL = "https://drive.usercontent.google.com/download?id=0B52yRXcdpG6MMnRNV3dTdGdYQ2M&export=download&confirm=t"
            print(f"Figer dataset: Downloading train.data.gz from {DATASET_URL}...")
            response = requests.get(DATASET_URL)
            if response.status_code == 200:
                with open(gzip_path, "wb") as f:
                    f.write(response.content)
                print("Figer dataset: Download complete.")
            else:
                raise Exception(f"Failed to download train.data.gz. Status code: {response.status_code}")

        print("Figer dataset: Extracting into train.data...")
        with gzip.open(gzip_path, "rb") as f_in:
            with open(data_path, "wb") as f_out:
                shutil.copyfileobj(f_in, f_out)
        os.remove(gzip_path)
        print("Figer dataset: Successfully extracted")

    return data_path


def load_or_preprocess(entries_coarse, entries_fine):
    coarse_serialized_path = os.path.join(base_dir, "figer-coarse.pickle")
    fine_serialized_path = os.path.join(base_dir, "figer-fine.pickle")

    # serialized already preprocessed data not found. Pre-process it not
    if not os.path.exists(coarse_serialized_path) or not os.path.exists(fine_serialized_path):
        if os.path.exists(coarse_serialized_path):
            os.remove(coarse_serialized_path)
        if os.path.exists(fine_serialized_path):
            os.remove(fine_serialized_path)

        data_path = download_and_unpack()

        # import Freebase types to figer labels mapper
        # this also makes sure that all tags in the mapper are the same as in the two lists above. This was not a given, I needed to do a lot of cleanup to get it there
        label_mapping: dict[str, tuple[str, str]] = {}
        with open(os.path.join(base_dir, "types.map"), "r") as f:
            coarse_labels_used = dict.fromkeys(coarse_labels, False)
            fine_labels_used = dict.fromkeys(fine_labels, False)
            for line in f:
                [old_label, new_label] = line.split("	", 1)
                labels_by_granularities = new_label.split("/")
                if len(labels_by_granularities) > 2:
                    new_label_tupel = (labels_by_granularities[1].strip(), labels_by_granularities[2].strip())
                else:
                    new_label_tupel = (labels_by_granularities[1].strip(), "")
                if new_label_tupel[0] not in (coarse_labels + [""]) or new_label_tupel[1] not in (fine_labels + [""]):
                    raise Exception(f"Unexpected label in label mapping: {new_label_tupel}")
                label_mapping[old_label.strip()] = new_label_tupel
                if new_label_tupel[0] != "":
                    coarse_labels_used[new_label_tupel[0]] = True
                if new_label_tupel[1] != "":
                    fine_labels_used[new_label_tupel[1]] = True

            # check if there are any labels that weren't covered by mapping
            for key, value in coarse_labels_used.items():
                if not value:
                    raise Exception(f"Coarse label {key} never occurred in mapping!")
            for key, value in fine_labels_used.items():
                if not value:
                    raise Exception(f"Fine label {key} never occurred in mapping!")

        # import dataset and apply mapping on it
        with open(data_path, "rb") as f:
            buf = f.read()
            n = 0
            i = 1
            while n < len(buf):
                msg_len, new_pos = _DecodeVarint32(buf, n)
                n = new_pos
                msg_buf = buf[n:n + msg_len]
                n += msg_len
                mention = entity_pb2.Mention()
                mention.ParseFromString(msg_buf)

                # concat sentence
                sentence = ""
                for token in mention.tokens:
                    if sentence != "" and token not in [".", ",", ":", ";"]:
                        sentence += " "
                    sentence += token

                # sometimes the entity names have additional remarks in brackets (since they are from wikipedia). Remove those
                entity_name = mention.entity_name.split("(", 1)[0].strip()

                coarse_label_set = set()
                fine_label_set = set()
                for label in mention.labels:
                    if new_label := label_mapping.get(label):
                        if new_label[0] != "":
                            coarse_label_set.add(new_label[0])
                        if new_label[1] != "":
                            fine_label_set.add(new_label[1])
                # only add entries that have both a coarse and a fine label
                # this way both coarse and fine datasets are identical except for the label
                if len(coarse_label_set) > 0 and len(fine_label_set) > 0:
                    coarse_label_list = list(map(lambda x: (entity_name, x), coarse_label_set))
                    entries_coarse.append((sentence, coarse_label_list))
                    fine_label_list = list(map(lambda x: (entity_name, x), fine_label_set))
                    entries_fine.append((sentence, fine_label_list))

                print(f"Figer dataset: Processed {i} entries", end="\r")
                i += 1
            print("")

        print(
            f"Figer dataset: Preprocessing completed. Writing result into {coarse_serialized_path} and {fine_serialized_path} now...")
        with open(coarse_serialized_path, "wb") as coarse_file:
            pickle.dump(entries_coarse, coarse_file)
        with open(fine_serialized_path, "wb") as fine_file:
            pickle.dump(entries_fine, fine_file)
        print()
        os.remove(data_path)
        print("Figer dataset: Writing complete.")

    # just load already preprocessed data from pickle file
    else:
        print(
            f"Figer dataset: Importing already preprocessed data from {coarse_serialized_path} and {fine_serialized_path} now...")
        with open(coarse_serialized_path, "rb") as coarse_file:
            entries_coarse += pickle.load(coarse_file)
        with open(fine_serialized_path, "rb") as fine_file:
            entries_fine += pickle.load(fine_file)
        print("Figer dataset: Imported preprocessed data from serialized file")


def get_labels_coarse():
    """Returns coarse labels"""
    return coarse_labels


def get_labels_fine():
    """Returns fine-grained labels"""
    return fine_labels


def get_annotated_sentences_coarse():
    """ """
    if len(entries_coarse) == 0:
        load_or_preprocess(entries_coarse, entries_fine)
        if len(entries_coarse) == 0:
            raise Exception("No fine entries found, preprocessing or loading must have failed!")
    return entries_coarse


def get_annotated_sentences_fine():
    """ """
    if len(entries_fine) == 0:
        load_or_preprocess(entries_coarse, entries_fine)
        if len(entries_fine) == 0:
            raise Exception("No fine entries found, preprocessing or loading must have failed!")
    return entries_fine
