import os
import json
import re
import requests


def download_json():
    """
    Downloads train.json from Hugging Face if it doesn't exist in the local directory.
    """
    base_dir = os.path.dirname(os.path.realpath(__file__))
    json_path = os.path.join(base_dir, "train.json")

    if not os.path.exists(json_path):
        DATASET_URL = "https://huggingface.co/datasets/Universal-NER/Pile-NER-type/resolve/main/train.json"
        print(f"PileNER - Downloading train.json from {DATASET_URL}...")
        response = requests.get(DATASET_URL)
        if response.status_code == 200:
            with open(json_path, "wb") as f:
                f.write(response.content)
            print("Download complete.")
        else:
            raise Exception(f"Failed to download train.json. Status code: {response.status_code}")
    return json_path


def get_labels():
    """
    Returns a list of unique entity labels (as strings) extracted from the questions in the conversations.
    It looks for human messages that ask questions of the form:
        "What describes <entity type> in the text?"
    """
    json_path = download_json()
    with open(json_path, "r", encoding="utf-8") as f:
        data = json.load(f)
    labels = set()  # Using a set to avoid duplicates

    for entry in data:
        for conversation in entry.get("conversations", []):
            if conversation["from"] == "human" and conversation["value"].startswith("What describes"):
                label = conversation["value"].replace("What describes ", "").replace(" in the text?", "").strip()
                labels.add(label.lower())  # Convert to lowercase before adding

    return sorted(labels)


def get_annotated_sentences():
    """
    Returns a list of annotated sentences.

    Each entry in the returned list is of the form:
       [[full_text], [(entity1, label1), (entity2, label2), ...]]

    The function assumes that:
      - The first human message starting with "Text:" contains the full text (after stripping the prefix).
      - Subsequent human questions in the conversation are in the form
            "What describes <entity type> in the text?"
        and are immediately followed by a GPT response that is a JSON-encoded list
        of entity mentions.
    """
    json_path = download_json()
    with open(json_path, "r", encoding="utf-8") as f:
        data = json.load(f)

    annotated_sentences = []

    for item in data:
        convs = item.get("conversations", [])
        full_text = None
        # Find the text from the first human turn that starts with "Text:"
        for conv in convs:
            if conv.get("from", "").lower() == "human" and conv.get("value", "").strip().lower().startswith("text:"):
                full_text = conv["value"][len("Text:"):].strip()
                break
        if full_text is None:
            continue  # Skip if no base text found

        annotations = []
        i = 0
        # Process conversation in order, looking for Q/A pairs.
        while i < len(convs):
            conv = convs[i]
            if conv.get("from", "").lower() == "human" and conv.get("value", "").strip().lower().startswith(
                    "what describes"):
                # Extract the entity type from the question.
                m = re.search(r"what describes\s+(.*?)\s+in the text\??", conv["value"].strip(), re.IGNORECASE)
                if m:
                    entity_type = m.group(1).strip()
                    # The next turn should be GPT's answer.
                    if i + 1 < len(convs) and convs[i + 1].get("from", "").lower() == "gpt":
                        answer_text = convs[i + 1].get("value", "").strip()
                        try:
                            # Parse the GPT answer, which should be a JSON-encoded list.
                            entities = json.loads(answer_text)
                            if isinstance(entities, list):
                                for ent in entities:
                                    annotations.append((ent, entity_type))
                        except json.JSONDecodeError:
                            # If parsing fails, skip the error.
                            pass
                    i += 2  # Skip the GPT answer
                else:
                    i += 1
            else:
                i += 1
        annotated_sentences.append([full_text, annotations])

    return annotated_sentences
