"""
This file provides a common interface for all the datasets used in this project.

Output:
For each dataset:

labels = ["object", "person", ...]

annotated_sentence = ("Barack Obama was the president of the United States.",[("Barack Obama", person), ("United States", location), ("United States", country)])
"""
import data.CoNLL.conll_interface as conll_interface
import data.Figer.figer_interface as figer_interface
import data.PileNERtype.pile_interface as pile_interface
import random


def get_labels(dataset):
    """
    Returns a list of all labels used in the specified dataset.
    """
    if dataset == "CoNLL":
        return conll_interface.get_labels()
    elif dataset == "Pile-NER-type":
        return pile_interface.get_labels()
    elif dataset == "FIGER-coarse":
        return figer_interface.get_labels_coarse()
    elif dataset == "FIGER-fine":
        return figer_interface.get_labels_fine()
    else:
        raise Exception


def get_annotated_sentences(dataset, test_instances=10, train_test_split=1.0, train=False):
    """
    Returns a list of annotated sentences reproducibly picked at random from the dataset.
    """
    random.seed(42)

    if dataset == "CoNLL":
        sentences = random.sample(conll_interface.get_annotated_sentences(),
                                  min(test_instances, len(conll_interface.get_annotated_sentences())))
    elif dataset == "Pile-NER-type":
        sentences = random.sample(pile_interface.get_annotated_sentences(),
                                  min(test_instances, len(pile_interface.get_annotated_sentences())))
    elif dataset == "FIGER-coarse":
        sentences = random.sample(figer_interface.get_annotated_sentences_coarse(),
                                  min(test_instances, len(figer_interface.get_annotated_sentences_coarse())))
    elif dataset == "FIGER-fine":
        sentences = random.sample(figer_interface.get_annotated_sentences_fine(),
                                  min(test_instances, len(figer_interface.get_annotated_sentences_fine())))
    else:
        raise Exception

    split_pos = max(0, min(len(sentences), int(train_test_split * len(sentences))))

    if train:
        return sentences[split_pos:]
    else:
        return sentences[:split_pos]


def get_label_dict(dataset, test_instances=10):
    """
    Returns a dictionary of example entities for each label.
    """
    label_dict = {}
    labels = get_labels(dataset)
    for label in labels:
        label_dict[label] = set()
    annotated_sentences = get_annotated_sentences(dataset, test_instances)
    for annotated_sentence in annotated_sentences:
        for labeled_entity in annotated_sentence[1]:
            if labeled_entity[1] not in label_dict:
                label_dict[labeled_entity[1]] = set()
            label_dict[labeled_entity[1]].add(labeled_entity[0])
    print(f"Label dict: {label_dict}")
    return label_dict
