"""
This file evaluates all NEC approaches on all datasets.
"""

import os
import csv
import datetime
import pandas as pd
from tqdm import tqdm
import data.data_manager as data_manager
from src.common_interface import classify_entity, set_label_dict


def run_NEC_tests(model_name, dataset, results_dir, test_instances=10):
    """
    Evaluates a model on a dataset on the NEC task for a chosen number of test instances and stores the
    results in the given directory
    """
    os.makedirs(results_dir, exist_ok=True)

    # Get current timestamp for unique file naming
    timestamp = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")

    # Define file paths with model name and timestamp in the file names
    csv_filename = os.path.join(results_dir, f"{timestamp}_{model_name}_{dataset}_metrics.csv")
    txt_filename = os.path.join(results_dir, f"{timestamp}_{model_name}_{dataset}_results.txt")

    labels = data_manager.get_labels(dataset)
    data = data_manager.get_annotated_sentences(dataset, test_instances)

    if model_name == "Word2Vec" or model_name == "T5-MLM-entity":
        label_dict = data_manager.get_label_dict(dataset, test_instances)
        print(label_dict)
        set_label_dict(model_name, label_dict)

    with open(csv_filename, mode="w", newline='', encoding="utf-8") as csv_file, \
            open(txt_filename, mode="w", encoding="utf-8") as txt_file:
        csv_writer = csv.writer(csv_file)
        csv_writer.writerow(
            ["Instance", "Model", "Dataset", "Sentence", "Entity", "True Annotation", "Predicted Annotation", "Correct"])

        for idx, entry in enumerate(tqdm(data, desc="Processing Instances", unit="instance")):
            sentence = entry[0]
            entity_dict = {}  # If there are multiple labels assigned to one entity,
            # it should count as correct if the model prediction is one of them.
            for entity in entry[1]:
                entity_name = entity[0]
                true_label = entity[1]

                if entity_name not in entity_dict:
                    entity_dict[entity_name] = set()
                entity_dict[entity_name].add(true_label)

            for entity_name, true_labels in entity_dict.items():
                predicted_label = classify_entity(model_name, sentence, entity_name, labels)
                correct = predicted_label in true_labels

                csv_writer.writerow(
                    [idx, model_name, dataset, sentence, entity_name, ", ".join(true_labels), predicted_label, correct])

                # Log results
                txt_file.write(f"\nInstance {idx}:")
                txt_file.write(f"\nSentence: {sentence}")
                txt_file.write(f"\nEntity: {entity_name}")
                txt_file.write(f"\nTrue Label(s): {', '.join(true_labels)}")
                txt_file.write(f"\nPredicted Label: {predicted_label}\n")
                txt_file.write("-" * 50 + "\n\n")

    print(f"Results saved to:\n  CSV: {csv_filename}\n  TXT: {txt_filename}")


def run_NEC_tests_all():
    models = ["DeepSeek-R1-Distill-Qwen-32B", "GLiNER", "Llama-3.1-8B", "T5-NLI", "T5-MLM-label", "T5-MLM-entity", "Word2Vec"]
    datasets = ["CoNLL", "FIGER-coarse", "FIGER-fine"]  # "Pile-NER-type"]
    for model in models:
        for dataset in datasets:
            print(f"Testing {model} on {dataset} dataset...")
            run_NEC_tests(model, dataset, "results", 100)


def read_NEC_metrics(directory):
    """
    Reads all CSV files in the given directory and prints the accuracy.
    """
    metrics = {}

    for filename in os.listdir(directory):
        if filename.endswith(".csv"):
            csv_filename = os.path.join(directory, filename)

            try:
                df = pd.read_csv(csv_filename)
                required_columns = {"Model", "Dataset", "Correct"}
                if not required_columns.issubset(df.columns):
                    print(f"Skipping {filename} due to missing columns.")
                    continue

                # Group by model and dataset, then compute accuracy
                for (model, dataset), group in df.groupby(["Model", "Dataset"]):
                    accuracy = group["Correct"].mean() * 100
                    metrics.setdefault((model, dataset), []).append(accuracy)

            except Exception as e:
                print(f"Error reading {filename}: {e}")

    print("\nModel Performance Summary:")
    for (model, dataset), accuracies in metrics.items():
        avg_accuracy = sum(accuracies) / len(accuracies)
        print(f"Model: {model}, Dataset: {dataset}, Accuracy: {avg_accuracy:.2f}%")


run_NEC_tests_all()
read_NEC_metrics("results")
