"""
This file defines and applies the NER pipeline with LLMs on chosen datasets and stores the results.
The results are evaluated / plotted in evaluation.py in the same folder.
"""

import re
import ast
from src.models.LLM_interface import LLM


ner_system_prompt = """You are part of a named entity recognition pipeline. For all provided labels, find all the fitting
entities in the target sentence as far as possible. Output your result as a list of pairs of strings ('entity', 'label')
Mark your result with <answer>Result</answer>.
Example:
Labels == ['person', 'organization', 'location', 'miscellaneous']
sentence: 'Europe rejects German call to boycott British lamb.' 
Result: <answer>[('Europe', 'organization'), ('Europe', 'location'), ('German', 'miscellaneous'), ('British', 'miscellaneous')]</answer>

If there is no valid entity in the target sentence, answer "<answer>[]</answer>".

Your Task:
"""

nli_system_prompt = """You are part of a named entity classification pipeline. 
Given an entity, find the best fitting label of the provided labels (do not invent your own labels!). 
Choose the label so that the sentence "Target entity is a chosen_label" makes the most sense.
Mark your result like this for easy extraction: <answer>predicted_class</answer>.

Example:
Labels == ['person', 'organization', 'location', 'miscellaneous']
Sentence: 'NASA sent astronauts to the moon.' 
Target Entity: NASA
Desired Result: <answer>organization</answer>

Your Task:
"""


def find_entities(model_name, sentence, labels):
    """Gets answer from model as list using the naive approach of asking the model to find all entities."""
    model = LLM.create(model_name)
    # We add the system prompt directly to the prompt, because DeepSeek documentation says that this is better.
    prompt = ner_system_prompt + "\nLabels: " + str(labels) + "\nTarget sentence: " + sentence
    answer = model(prompt)
    answer = answer.encode("utf-8").decode("utf-8-sig")
    # Extract content between <answer> tags
    match = re.search(r"<answer>(.*?)</answer>", answer, re.DOTALL)
    if match:
        extracted_text = match.group(1).strip()
        extracted_text = extracted_text.replace("\n", "").replace("  ", " ")
        try:
            # Evaluate as a Python list of tuples
            result = ast.literal_eval(extracted_text)
            if isinstance(result, list) and all(isinstance(i, tuple) and len(i) == 2 for i in result):
                return result
        except Exception as e:
            print(e)
    return "No answer was marked, here is the model's response: " + answer


def classify_entity(model_name, sentence, entity, labels):
    """
    Uses NLI to classify a specified entity in a sentence. Returns string entity class.
    """
    model = LLM.create(model_name)
    # We add the system prompt directly to the prompt, because DeepSeek documentation says that this is better.
    prompt = nli_system_prompt + "\nLabels: " + str(labels) + "\nTarget sentence: " + sentence + "\nTarget entity: " + entity
    answer = model(prompt)
    answer = answer.encode("utf-8").decode("utf-8-sig")
    # Extract content between <answer> tags
    match = re.search(r"<answer>(.*?)</answer>", answer, re.DOTALL)
    if match:
        result = match.group(1).strip()
        return result
    return "No answer was marked, here is the model's response: " + answer
