"""
This file contains functions calculating precision, recall and F1 score for NER.
Input format: two lists like this: [('entity', 'label'), ('entity', 'label'), ('entity', 'label')]]
"""
import os
import csv
import datetime
import pandas as pd
from tqdm import tqdm
import data.data_manager as data_manager
from src.common_interface import find_entities


def precision(true, predicted):
    true_set = set(true)
    predicted_set = set(predicted)

    if not predicted_set and not true_set:
        return 1.0
    if not predicted_set or not true_set:
        return 0.0

    tp = len(true_set & predicted_set)
    return tp / len(predicted_set)


def recall(true, predicted):
    true_set = set(true)
    predicted_set = set(predicted)

    if not predicted_set and not true_set:
        return 1.0
    if not predicted_set or not true_set:
        return 0.0

    tp = len(true_set & predicted_set)
    return tp / len(true_set)


def f1_score(true, predicted):
    p = precision(true, predicted)
    r = recall(true, predicted)

    if p + r == 0:
        return 0.0

    return 2 * (p * r) / (p + r)


def NER_metrics(model_name, dataset, results_dir, test_instances=10):
    """
    Evaluates a model on a dataset on the NER task for a chosen number of test instances and stores the
    results in the given directory
    """
    os.makedirs(results_dir, exist_ok=True)

    # Get current timestamp for unique file naming
    timestamp = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")

    # Define file paths with model name and timestamp in the file names
    csv_filename = os.path.join(results_dir, f"{timestamp}_{model_name}_metrics.csv")
    txt_filename = os.path.join(results_dir, f"{timestamp}_{model_name}_results.txt")

    labels = data_manager.get_labels(dataset)
    data = data_manager.get_annotated_sentences(dataset, test_instances)

    with open(csv_filename, mode="w", newline='', encoding="utf-8") as csv_file, \
            open(txt_filename, mode="w", encoding="utf-8") as txt_file:
        csv_writer = csv.writer(csv_file)
        csv_writer.writerow(
            ["Instance", "Dataset", "Sentence", "True Annotation", "Predicted Annotation", "Precision", "Recall", "F1 Score"])

        for idx, entry in enumerate(tqdm(data, desc="Processing Instances", unit="instance")):
            sentence = entry[0]
            true_annotation = entry[1]

            predicted_annotation = find_entities(model_name, sentence, labels)

            prec = precision(true_annotation, predicted_annotation)
            reca = recall(true_annotation, predicted_annotation)
            f1 = f1_score(true_annotation, predicted_annotation)

            csv_writer.writerow([idx, dataset, sentence, true_annotation, predicted_annotation, prec, reca, f1])

            # Log stuff
            txt_file.write(f"Instance {idx}:\n")
            txt_file.write(f"Sentence: {sentence}\n")
            txt_file.write(f"True Annotation: {true_annotation}\n")
            txt_file.write(f"Predicted Annotation: {predicted_annotation}\n")
            txt_file.write(f"Precision: {prec}; Recall: {reca}; F1-score: {f1} \n")
            txt_file.write("-" * 50 + "\n\n")

    print(f"Results saved to:\n  CSV: {csv_filename}\n  TXT: {txt_filename}")


def read_NER_metrics(directory):
    """
    Reads all CSV files in the given directory and prints the mean precision, recall, and F1 score.
    """
    for filename in os.listdir(directory):
        if filename.endswith(".csv"):  # Ensure we only process CSV files
            csv_filename = os.path.join(directory, filename)

            try:
                df = pd.read_csv(csv_filename)

                if {"Precision", "Recall", "F1 Score"}.issubset(df.columns):
                    mean_precision = df["Precision"].mean()
                    mean_recall = df["Recall"].mean()
                    mean_f1_score = df["F1 Score"].mean()

                    print(f"Results from: {filename}")
                    print(f"Mean Precision: {mean_precision:.4f}")
                    print(f"Mean Recall: {mean_recall:.4f}")
                    print(f"Mean F1 Score: {mean_f1_score:.4f}")
                    print("-" * 50)
                else:
                    print(f"Skipping {filename}: Missing required columns.")
            except Exception as e:
                print(f"Error reading {filename}: {e}")
