"""
This module implements the entity masking approach for the named entity recogition task. It uses the T5 model and allows for finetuning of the model.
"""

import random
import numpy as np
from transformers import T5ForConditionalGeneration, T5Tokenizer, Trainer, TrainingArguments, DataCollatorForSeq2Seq
from datasets import Dataset, DatasetDict

model_name = "google-t5/t5-base"


def load_base():
    global model
    global tokenizer
    print("Loading model: T5 MLM entity")
    tokenizer = T5Tokenizer.from_pretrained(model_name)
    model = T5ForConditionalGeneration.from_pretrained(model_name)
    print("Finished loading model: T5 MLM entity")


def load_finetuned(input_dir):
    global model
    global tokenizer
    print(f"Loading model: T5 MLM entity finetuned ({input_dir})")
    tokenizer = T5Tokenizer.from_pretrained(input_dir)
    model = T5ForConditionalGeneration.from_pretrained(input_dir)
    print(f"Finished loading model: T5 MLM entity finetuned")


def set_label_dict(label_dict):
    global label_representatives
    label_representatives = label_dict


def classify_entity(sentence, entity, labels):
    global label_representatives

    masked_sentence = sentence.replace(entity, "<extra_id_0>", 1)
    input_ids = tokenizer(masked_sentence, return_tensors="pt").input_ids

    best_label = None
    best_label_loss = float("inf")

    for label in labels:
        best_rep_loss = float("inf")
        representatives = label_representatives[label]

        for representative in representatives:
            # do not cheat if entity is one of the representatives already
            if entity == representative:
                continue

            representative_ids = tokenizer(f"<extra_id_0> {representative} <extra_id_1>", return_tensors="pt").input_ids
            loss = model(input_ids=input_ids, labels=representative_ids).loss.item()
            best_rep_loss = min(best_rep_loss, loss)

        if best_rep_loss < best_label_loss:
            best_label_loss = best_rep_loss
            best_label = label

    return best_label if best_label else labels[0]


def finetune_model(sentences, entities, labels, output_dir, epochs=10):
    input_texts = []
    target_texts = []

    random.seed(42)

    for i in range(len(sentences)):
        sentence = sentences[i]
        entity = entities[i]
        label = labels[i]
        representatives = label_representatives[label]
        representative = random.choice(list(representatives))
        input_texts.append(sentence.replace(entity, "<extra_id_0>", 1))
        target_texts.append(f"<extra_id_0> {representative} <extra_id_1>")

    model_input = tokenizer(input_texts, return_tensors="pt", padding="max_length", truncation=True, max_length=128)
    targets = tokenizer(target_texts, return_tensors="pt", padding="max_length", truncation=True, max_length=128)

    model_input["input_ids"] = np.array(model_input["input_ids"])
    model_input["attention_mask"] = np.array(model_input["attention_mask"])
    model_input["labels"] = np.array(targets["input_ids"])

    dataset = Dataset.from_dict({
        "input_ids": model_input["input_ids"],
        "attention_mask": model_input["attention_mask"],
        "labels": model_input["labels"]
    })

    print(dataset)

    # split into training and validation data (20% used for validation)
    train_test_split = dataset.train_test_split(test_size=0.2, shuffle=True, seed=0)

    dataset = DatasetDict({
        "train": train_test_split["train"],
        "validation": train_test_split["test"]
    })

    dataset.set_format(type="torch", columns=["input_ids", "attention_mask", "labels"])

    training_args = TrainingArguments(
        output_dir="./src/models/t5_mlm_entity_finetuned_model/checkpoints/",
        eval_strategy="epoch",
        learning_rate=5e-5,
        per_device_train_batch_size=8,
        per_device_eval_batch_size=8,
        num_train_epochs=epochs,
        weight_decay=0.01,
        save_strategy="epoch",
        push_to_hub=False,
        logging_dir="./logs",
    )

    data_collator = DataCollatorForSeq2Seq(tokenizer, model=model)

    trainer = Trainer(
        model=model,
        args=training_args,
        train_dataset=dataset["train"],
        eval_dataset=dataset["validation"],
        tokenizer=tokenizer,
        data_collator=data_collator,
    )

    trainer.train()

    model.save_pretrained(output_dir)
    tokenizer.save_pretrained(output_dir)


load_base()
# load_finetuned("./src/models/t5_mlm_entity_finetuned_model/checkpoints/checkpoint-12200")
