"""
This module implements the natural language inference approach for the named entity recogition task. It uses the T5 model and allows for finetuning of the model.
"""

import torch
from torch.nn.functional import softmax
from transformers import T5ForConditionalGeneration, T5Tokenizer, Trainer, TrainingArguments, DataCollatorForSeq2Seq
from datasets import Dataset, DatasetDict

label_map = {True: "entailment", False: "contradiction"}

# Use t5-base for testing because it is smaller
model_name = "google-t5/t5-base"


# model_name = "google/t5_xxl_true_nli_mixture"

def load_base():
    global model
    global tokenizer
    print("Loading model: T5 NLI")
    tokenizer = T5Tokenizer.from_pretrained(model_name)
    model = T5ForConditionalGeneration.from_pretrained(model_name)
    print("Finished loading model: T5 NLI")


def load_finetuned(input_dir):
    global model
    global tokenizer
    print(f"Loading model: T5 NLI finetuned ({input_dir})")
    tokenizer = T5Tokenizer.from_pretrained(input_dir)
    model = T5ForConditionalGeneration.from_pretrained(input_dir)
    print(f"Finished loading model: T5 NLI finetuned")


def infer_nli(premise, hypothesis):
    input_text = f"nli hypothesis: {hypothesis} premise: {premise}"
    inputs = tokenizer(input_text, return_tensors="pt")
    output_ids = model.generate(**inputs)
    result = tokenizer.decode(output_ids[0], skip_special_tokens=True)

    return result


def infer_nli_probabilities(premise, hypothesis):
    input_text = f"nli hypothesis: {hypothesis} premise: {premise}"

    inputs = tokenizer(input_text, return_tensors="pt")

    decoder_input_ids = torch.tensor([[tokenizer.pad_token_id]])

    with torch.no_grad():
        outputs = model(**inputs, decoder_input_ids=decoder_input_ids)

    logits = outputs.logits

    # Use this code if we also want to get the classification:
    # predicted_ids = torch.argmax(logits, dim=-1)
    # predicted_label = tokenizer.decode(predicted_ids[0], skip_special_tokens=True)

    probabilities = softmax(logits, dim=-1)

    label_tokens = ["entailment", "contradiction"]
    label_ids = [tokenizer.encode(label, add_special_tokens=False)[0] for label in label_tokens]

    label_probs = {label: probabilities[0, 0, token_id].item() for label, token_id in zip(label_tokens, label_ids)}

    return label_probs


def classify_entity(sentence, entity, labels):
    best_label = None
    best_prob = 0
    # Find label with highest probability of entailment
    for label in labels:
        hypothesis = f"{entity} is {label}"
        label_probs = infer_nli_probabilities(sentence, hypothesis)

        # print(f"label: {label}, entailment prob: {label_probs["entailment"]}, contradiction prob: {label_probs["contradiction"]}")

        entailment_prob = label_probs["entailment"]

        if entailment_prob > best_prob:
            best_prob = entailment_prob
            best_label = label

    return best_label if best_label else labels[0]


def preprocess_data(sample):
    input_text = f"nli hypothesis: {sample['hypothesis']} premise: {sample['premise']}"
    target_text = label_map[bool(sample['entailment'])]

    tokenized_input = tokenizer(input_text, padding="max_length", truncation=True, max_length=128)
    tokenized_target = tokenizer(target_text, padding="max_length", truncation=True, max_length=10)

    tokenized_input["labels"] = tokenized_target["input_ids"]
    return tokenized_input


def finetune_model(premises, hypotheses, entailment, output_dir, epochs=10):
    # TODO: should we use dataset on a higher level as well?
    data_dict = {"premise": premises, "hypothesis": hypotheses, "entailment": entailment}
    dataset = Dataset.from_dict(data_dict)

    print(dataset)

    # split into training and validation data (20% used for validation)
    train_test_split = dataset.train_test_split(test_size=0.2, shuffle=True, seed=0)

    dataset = DatasetDict({
        "train": train_test_split["train"],
        "validation": train_test_split["test"]
    })

    # preprocess for NLI task
    tokenized_datasets = dataset.map(preprocess_data, batched=False)
    tokenized_datasets.set_format(type="torch", columns=["input_ids", "attention_mask", "labels"])

    training_args = TrainingArguments(
        output_dir="./src/models/t5_nli_finetuned_model/checkpoints/",
        eval_strategy="epoch",
        learning_rate=5e-5,
        per_device_train_batch_size=8,
        per_device_eval_batch_size=8,
        num_train_epochs=epochs,
        weight_decay=0.01,
        save_strategy="epoch",
        push_to_hub=False,
        logging_dir="./logs",
    )

    data_collator = DataCollatorForSeq2Seq(tokenizer, model=model)

    trainer = Trainer(
        model=model,
        args=training_args,
        train_dataset=tokenized_datasets["train"],
        eval_dataset=tokenized_datasets["validation"],
        tokenizer=tokenizer,
        data_collator=data_collator,
    )

    trainer.train()

    model.save_pretrained(output_dir)
    tokenizer.save_pretrained(output_dir)


load_base()
# load_finetuned("./src/models/t5_nli_finetuned_model/checkpoints/checkpoint-85500")
